@@ -61,6 +61,7 @@ import com.google.zxing.integration.android.IntentIntegrator;
 import com.google.zxing.integration.android.IntentResult;
 
 import java.lang.ref.WeakReference;
+import java.util.List;
 
 import javax.inject.Inject;
 
@@ -77,21 +78,25 @@ import cx.ring.model.Account;
 import cx.ring.model.CallContact;
 import cx.ring.model.Conference;
 import cx.ring.model.Conversation;
+import cx.ring.model.DaemonEvent;
 import cx.ring.model.Phone;
 import cx.ring.model.Uri;
 import cx.ring.service.LocalService;
-import cx.ring.services.StateService;
+import cx.ring.services.AccountService;
+import cx.ring.services.CallService;
 import cx.ring.utils.ActionHelper;
 import cx.ring.utils.BlockchainInputHandler;
 import cx.ring.utils.ClipboardHelper;
 import cx.ring.utils.ContentUriHandler;
+import cx.ring.utils.Observable;
+import cx.ring.utils.Observer;
 
 public class SmartListFragment extends Fragment implements SearchView.OnQueryTextListener,
         HomeActivity.Refreshable,
         SmartListAdapter.SmartListAdapterCallback,
         Conversation.ConversationActionCallback,
         ClipboardHelper.ClipboardHelperCallback,
-        LocalService.NameLookupCallback {
+        Observer<DaemonEvent> {
     private static final String TAG = SmartListFragment.class.getSimpleName();
 
     private static final int USER_INPUT_DELAY = 300;
@@ -107,6 +112,8 @@ public class SmartListFragment extends Fragment implements SearchView.OnQueryTex
     private SearchView mSearchView = null;
     private MenuItem mSearchMenuItem = null;
     private MenuItem mDialpadMenuItem = null;
+    private String mLastBlockchainQuery = null;
+    private Boolean isSearching = false;
 
     @BindView(R.id.confs_list)
     ListView mList;
@@ -132,7 +139,10 @@ public class SmartListFragment extends Fragment implements SearchView.OnQueryTex
     private Handler mUserInputHandler;
 
     @Inject
-    StateService mStateService;
+    AccountService mAccountService;
+
+    @Inject
+    CallService mCallService;
 
     final BroadcastReceiver receiver = new BroadcastReceiver() {
         @Override
@@ -148,6 +158,41 @@ public class SmartListFragment extends Fragment implements SearchView.OnQueryTex
     public static final int REQUEST_TRANSFER = 10;
     public static final int REQUEST_CONF = 20;
 
+    private Observer<DaemonEvent> mRinguifyObserver = new Observer<DaemonEvent>() {
+
+        private void updateContactRingId(String name, String address) {
+
+            if (TextUtils.isEmpty(name) || TextUtils.isEmpty(address)) {
+                return;
+            }
+
+            LocalService service = mCallbacks.getService();
+            service.updateConversationContactWithRingId(name, address);
+            RingApplication.uiHandler.post(new Runnable() {
+                @Override
+                public void run() {
+                    mSmartListAdapter.notifyDataSetChanged();
+                }
+            });
+        }
+
+        @Override
+        public void update(Observable observable, DaemonEvent event) {
+            if (event == null) {
+                return;
+            }
+
+            if (event.getEventType() == DaemonEvent.EventType.REGISTERED_NAME_FOUND) {
+                String name = event.getEventInput(DaemonEvent.EventInput.NAME, String.class);
+                String address = event.getEventInput(DaemonEvent.EventInput.ADDRESS, String.class);
+                int state = event.getEventInput(DaemonEvent.EventInput.STATE, Integer.class);
+                if (state == 0) {
+                    updateContactRingId(name, address);
+                }
+            }
+        }
+    };
+
     @Override
     public void onAttach(Activity activity) {
         Log.d(TAG, "onAttach");
@@ -161,6 +206,7 @@ public class SmartListFragment extends Fragment implements SearchView.OnQueryTex
     }
 
     public void refresh() {
+
         LocalService service = mCallbacks.getService();
         if (service == null) {
             Log.e(TAG, "refresh: null service");
@@ -180,6 +226,38 @@ public class SmartListFragment extends Fragment implements SearchView.OnQueryTex
         } else {
             this.presentNetworkErrorPanel(service);
         }
+
+        // If conversation has been created based on a blockchained discovered username
+        // we update the contact with its RingId
+        searchForRingIdInBlockchain();
+    }
+
+    private void searchForRingIdInBlockchain() {
+        LocalService service = mCallbacks.getService();
+        if (service == null || !service.isConnected() || !service.areConversationsLoaded()) {
+            return;
+        }
+        List<Conversation> conversations = service.getConversations();
+        for (Conversation conversation : conversations) {
+            CallContact contact = conversation.getContact();
+            if (contact == null) {
+                continue;
+            }
+
+            Uri contactUri = new Uri(contact.getIds().get(0));
+            if (contactUri.isRingId()) {
+                return;
+            }
+
+            if (contact.getPhones().isEmpty()) {
+                mAccountService.lookupName("", "", contact.getDisplayName());
+            } else {
+                Phone phone = contact.getPhones().get(0);
+                if (!phone.getNumber().isRingId()) {
+                    mAccountService.lookupName("", "", contact.getDisplayName());
+                }
+            }
+        }
     }
 
     @Override
@@ -196,7 +274,6 @@ public class SmartListFragment extends Fragment implements SearchView.OnQueryTex
         IntentFilter intentFilter = new IntentFilter();
         intentFilter.addAction(LocalService.ACTION_CONF_UPDATE);
         intentFilter.addAction(LocalService.ACTION_CONF_LOADED);
-        intentFilter.addAction(LocalService.ACTION_ACCOUNT_UPDATE);
         getActivity().registerReceiver(receiver, intentFilter);
     }
 
@@ -210,6 +287,8 @@ public class SmartListFragment extends Fragment implements SearchView.OnQueryTex
     @Override
     public void onResume() {
         super.onResume();
+        mAccountService.addObserver(this);
+        mAccountService.addObserver(mRinguifyObserver);
         Log.d(TAG, "onResume");
         ((HomeActivity) getActivity()).setToolbarState(false, R.string.app_name);
         refresh();
@@ -227,6 +306,7 @@ public class SmartListFragment extends Fragment implements SearchView.OnQueryTex
                 displayFloatingActionButtonWithDelay(true, 50);
                 setOverflowMenuVisible(menu, true);
                 setLoading(false);
+                isSearching = false;
                 return true;
             }
 
@@ -236,6 +316,7 @@ public class SmartListFragment extends Fragment implements SearchView.OnQueryTex
                 displayFloatingActionButtonWithDelay(false, 0);
                 setOverflowMenuVisible(menu, false);
                 setLoading(false);
+                isSearching = true;
                 return true;
             }
         });
@@ -246,18 +327,20 @@ public class SmartListFragment extends Fragment implements SearchView.OnQueryTex
         mSearchView.setLayoutParams(new Toolbar.LayoutParams(Toolbar.LayoutParams.MATCH_PARENT, Toolbar.LayoutParams.MATCH_PARENT));
         mSearchView.setImeOptions(EditorInfo.IME_ACTION_GO);
 
-        Intent i = getActivity().getIntent();
-        switch (i.getAction()) {
-            case Intent.ACTION_VIEW:
-            case Intent.ACTION_CALL:
-                mSearchView.setQuery(i.getDataString(), true);
-                break;
-            case Intent.ACTION_DIAL:
-                mSearchMenuItem.expandActionView();
-                mSearchView.setQuery(i.getDataString(), false);
-                break;
-            default:
-                break;
+        Intent intent = getActivity().getIntent();
+        if (intent != null && intent.getAction() != null) {
+            switch (intent.getAction()) {
+                case Intent.ACTION_VIEW:
+                case Intent.ACTION_CALL:
+                    mSearchView.setQuery(intent.getDataString(), true);
+                    break;
+                case Intent.ACTION_DIAL:
+                    mSearchMenuItem.expandActionView();
+                    mSearchView.setQuery(intent.getDataString(), false);
+                    break;
+                default:
+                    break;
+            }
         }
     }
 
@@ -304,7 +387,7 @@ public class SmartListFragment extends Fragment implements SearchView.OnQueryTex
         if (TextUtils.isEmpty(query)) {
             mNewContact.setVisibility(View.GONE);
         } else {
-            Account currentAccount = mStateService.getCurrentAccount();
+            Account currentAccount = mAccountService.getCurrentAccount();
             if (currentAccount == null) {
                 Log.w(TAG, "No account selected");
                 return false;
@@ -324,15 +407,16 @@ public class SmartListFragment extends Fragment implements SearchView.OnQueryTex
 
                 // Ring search
                 if (mBlockchainInputHandler == null) {
-                    mBlockchainInputHandler = new BlockchainInputHandler(new WeakReference<>(mCallbacks.getService()), this);
+                    mBlockchainInputHandler = new BlockchainInputHandler(new WeakReference<>(mAccountService));
                 }
 
                 // searching for a ringId or a blockchained username
                 if (!mBlockchainInputHandler.isAlive()) {
-                    mBlockchainInputHandler = new BlockchainInputHandler(new WeakReference<>(mCallbacks.getService()), this);
+                    mBlockchainInputHandler = new BlockchainInputHandler(new WeakReference<>(mAccountService));
                 }
 
                 mBlockchainInputHandler.enqueueNextLookup(query);
+                mLastBlockchainQuery = query;
             }
         }
 
@@ -387,6 +471,13 @@ public class SmartListFragment extends Fragment implements SearchView.OnQueryTex
         return inflatedView;
     }
 
+    @Override
+    public void onPause() {
+        super.onPause();
+        mAccountService.removeObserver(this);
+        mAccountService.removeObserver(mRinguifyObserver);
+    }
+
     @OnClick(R.id.newcontact_element)
     void newContactClicked(View v) {
         CallContact c = (CallContact) v.getTag();
@@ -429,7 +520,7 @@ public class SmartListFragment extends Fragment implements SearchView.OnQueryTex
 
         // We add the contact to the current State so that we can
         // get it from whatever part of the app as "an already used contact"
-        mStateService.addContact(c);
+        mCallService.addContact(c);
 
         Intent intent = new Intent()
                 .setClass(getActivity(), ConversationActivity.class)
@@ -718,11 +809,6 @@ public class SmartListFragment extends Fragment implements SearchView.OnQueryTex
         }
     }
 
-    @Override
-    public void onFound(String name, String address) {
-        displayNewContactRowWithName(name, address);
-    }
-
     private void displayNewContactRowWithName(String name, String address) {
         ((TextView) mNewContact.findViewById(R.id.display_name)).setText(name);
         CallContact contact = CallContact.buildUnknown(name, address);
@@ -731,22 +817,53 @@ public class SmartListFragment extends Fragment implements SearchView.OnQueryTex
     }
 
     @Override
-    public void onInvalidName(String name) {
-        Uri uri = new Uri(name);
-        if (uri.isRingId()) {
-            displayNewContactRowWithName(name, null);
-        } else {
-            mNewContact.setVisibility(View.GONE);
+    public void update(Observable observable, final DaemonEvent event) {
+        if (event == null) {
+            return;
         }
-    }
 
-    @Override
-    public void onError(String name, String address) {
-        Uri uri = new Uri(address);
-        if (uri.isRingId()) {
-            displayNewContactRowWithName(name, address);
-        } else {
-            mNewContact.setVisibility(View.GONE);
+        if (event.getEventType() == DaemonEvent.EventType.REGISTERED_NAME_FOUND) {
+            if (!isSearching) {
+                return;
+            }
+            RingApplication.uiHandler.post(new Runnable() {
+                @Override
+                public void run() {
+                    String name = event.getEventInput(DaemonEvent.EventInput.NAME, String.class);
+                    String address = event.getEventInput(DaemonEvent.EventInput.ADDRESS, String.class);
+
+                    int state = event.getEventInput(DaemonEvent.EventInput.STATE, Integer.class);
+                    switch (state) {
+                        case 0:
+                            // on found
+                            if (!TextUtils.isEmpty(mLastBlockchainQuery) && mLastBlockchainQuery.equals(name)) {
+                                displayNewContactRowWithName(name, address);
+                                mLastBlockchainQuery = null;
+                            } else {
+                                mNewContact.setVisibility(View.GONE);
+                            }
+                            break;
+                        case 1:
+                            // invalid name
+                            Uri uriName = new Uri(name);
+                            if (uriName.isRingId()) {
+                                displayNewContactRowWithName(name, null);
+                            } else {
+                                mNewContact.setVisibility(View.GONE);
+                            }
+                            break;
+                        default:
+                            // on error
+                            Uri uriAddress = new Uri(address);
+                            if (uriAddress.isRingId()) {
+                                displayNewContactRowWithName(name, address);
+                            } else {
+                                mNewContact.setVisibility(View.GONE);
+                            }
+                            break;
+                    }
+                }
+            });
         }
     }
 }
