@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2004-2016 Savoir-faire Linux Inc.
+ *  Copyright (C) 2004-2017 Savoir-faire Linux Inc.
  *
  *  Author: Guillaume Roguez <guillaume.roguez@savoirfairelinux.com>
  *
@@ -26,8 +26,10 @@
 #include "upnp/upnp_control.h"
 
 #include <pjlib.h>
+#include <msgpack.hpp>
 
 #include <utility>
+#include <tuple>
 #include <algorithm>
 #include <sstream>
 #include <chrono>
@@ -41,6 +43,8 @@
 
 namespace ring {
 
+static constexpr unsigned STUN_MAX_PACKET_SIZE {8192};
+
 // TODO: C++14 ? remove me and use std::min
 template< class T >
 static constexpr const T& min( const T& a, const T& b ) {
@@ -52,7 +56,7 @@ register_thread()
 {
     // We have to register the external thread so it could access the pjsip frameworks
     if (!pj_thread_is_registered()) {
-#if __GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 8)
+#if __GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 8) || defined RING_UWP
         static thread_local pj_thread_desc desc;
         static thread_local pj_thread_t *this_thread;
 #else
@@ -64,6 +68,83 @@ register_thread()
     }
 }
 
+/**
+ * Add stun/turn servers or default host as candidates
+ */
+static void
+add_stun_server(pj_ice_strans_cfg& cfg, int af)
+{
+    if (cfg.stun_tp_cnt >= PJ_ICE_MAX_STUN)
+        throw std::runtime_error("Too many STUN servers");
+    auto& stun = cfg.stun_tp[cfg.stun_tp_cnt++];
+
+    pj_ice_strans_stun_cfg_default(&stun);
+    stun.cfg.max_pkt_size = STUN_MAX_PACKET_SIZE;
+    stun.af = af;
+
+    RING_DBG("[ice] added host stun server");
+}
+
+static void
+add_stun_server(pj_pool_t& pool, pj_ice_strans_cfg& cfg, const StunServerInfo& info)
+{
+    if (cfg.stun_tp_cnt >= PJ_ICE_MAX_STUN)
+        throw std::runtime_error("Too many STUN servers");
+
+    IpAddr ip {info.uri};
+
+    // Given URI cannot be DNS resolved or not IPv4 or IPv6?
+    // This prevents a crash into PJSIP when ip.toString() is called.
+    if (ip.getFamily() == AF_UNSPEC) {
+        RING_WARN("[ice] STUN server '%s' not used, unresolvable address", info.uri.c_str());
+        return;
+    }
+
+    auto& stun = cfg.stun_tp[cfg.stun_tp_cnt++];
+    pj_ice_strans_stun_cfg_default(&stun);
+    pj_strdup2_with_null(&pool, &stun.server, ip.toString().c_str());
+    stun.af = ip.getFamily();
+    stun.port = PJ_STUN_PORT;
+    stun.cfg.max_pkt_size = STUN_MAX_PACKET_SIZE;
+
+    RING_DBG("[ice] added stun server '%s', port %d", pj_strbuf(&stun.server), stun.port);
+}
+
+static void
+add_turn_server(pj_pool_t& pool, pj_ice_strans_cfg& cfg, const TurnServerInfo& info)
+{
+    if (cfg.turn_tp_cnt >= PJ_ICE_MAX_TURN)
+        throw std::runtime_error("Too many TURN servers");
+
+    IpAddr ip {info.uri};
+
+    // Same comment as add_stun_server()
+    if (ip.getFamily() == AF_UNSPEC) {
+        RING_WARN("[ice] TURN server '%s' not used, unresolvable address", info.uri.c_str());
+        return;
+    }
+
+    auto& turn = cfg.turn_tp[cfg.turn_tp_cnt++];
+    pj_ice_strans_turn_cfg_default(&turn);
+    pj_strdup2_with_null(&pool, &turn.server, ip.toString().c_str());
+    turn.af = ip.getFamily();
+    turn.port = PJ_STUN_PORT;
+    turn.cfg.max_pkt_size = STUN_MAX_PACKET_SIZE;
+
+    // Authorization (only static plain password supported yet)
+    if (not info.password.empty()) {
+        turn.auth_cred.type = PJ_STUN_AUTH_CRED_STATIC;
+        turn.auth_cred.data.static_cred.data_type = PJ_STUN_PASSWD_PLAIN;
+        pj_cstr(&turn.auth_cred.data.static_cred.realm, info.realm.c_str());
+        pj_cstr(&turn.auth_cred.data.static_cred.username, info.username.c_str());
+        pj_cstr(&turn.auth_cred.data.static_cred.data, info.password.c_str());
+    }
+
+    RING_DBG("[ice] added turn server '%s', port %d", pj_strbuf(&turn.server), turn.port);
+}
+
+//##################################################################################################
+
 IceTransport::Packet::Packet(void *pkt, pj_size_t size)
     : data(new char[size]), datalen(size)
 {
@@ -109,7 +190,7 @@ IceTransport::IceTransport(const char* name, int component_count, bool master,
         upnp_.reset(new upnp::Controller());
 
     auto& iceTransportFactory = Manager::instance().getIceTransportFactory();
-    config_ = iceTransportFactory.getIceCfg();
+    config_ = iceTransportFactory.getIceCfg(); // config copy
 
     pool_.reset(pj_pool_create(iceTransportFactory.getPoolFactory(),
                                "IceTransport.pool", 512, 512, NULL));
@@ -121,50 +202,13 @@ IceTransport::IceTransport(const char* name, int component_count, bool master,
     icecb.on_rx_data = cb_on_rx_data;
     icecb.on_ice_complete = cb_on_ice_complete;
 
-    /* STUN */
-    if (not options.stunServer.empty()) {
-        const auto n = options.stunServer.rfind(':');
-        if (n != std::string::npos) {
-            const auto p = options.stunServer.c_str();
-            pj_strset(&config_.stun.server, (char*)p, n);
-            config_.stun.port = (pj_uint16_t)std::atoi(p+n+1);
-        } else {
-            pj_cstr(&config_.stun.server, options.stunServer.c_str());
-            config_.stun.port = PJ_STUN_PORT;
-        }
-        RING_WARN("ICE: STUN='%s', PORT=%d", options.stunServer.c_str(),
-                 config_.stun.port);
-    } else
-        config_.stun.port = 0;
+    // Add STUN servers
+    for (auto& server : options.stunServers)
+        add_stun_server(*pool_, config_, server);
 
-    /* TURN */
-    if (not options.turnServer.empty()) {
-        const auto n = options.turnServer.rfind(':');
-        if (n != std::string::npos) {
-            const auto p = options.turnServer.c_str();
-            pj_strset(&config_.turn.server, (char*)p, n);
-            config_.turn.port = (pj_uint16_t)std::atoi(p+n+1);
-        } else {
-            pj_cstr(&config_.turn.server, options.turnServer.c_str());
-            config_.turn.port = PJ_STUN_PORT;
-        }
-
-        // Authorization (only static plain password supported yet)
-        if (not options.turnServerPwd.empty()) {
-            config_.turn.auth_cred.type = PJ_STUN_AUTH_CRED_STATIC;
-            config_.turn.auth_cred.data.static_cred.data_type = PJ_STUN_PASSWD_PLAIN;
-            pj_cstr(&config_.turn.auth_cred.data.static_cred.realm, options.turnServerRealm.c_str());
-            pj_cstr(&config_.turn.auth_cred.data.static_cred.username, options.turnServerUserName.c_str());
-            pj_cstr(&config_.turn.auth_cred.data.static_cred.data, options.turnServerPwd.c_str());
-        }
-
-        // Only UDP yet
-        config_.turn.conn_type = PJ_TURN_TP_UDP;
-
-        RING_WARN("ICE: TURN='%s', PORT=%d", options.turnServer.c_str(),
-                 config_.turn.port);
-    } else
-        config_.turn.port = 0;
+    // Add TURN servers
+    for (auto& server : options.turnServers)
+        add_turn_server(*pool_, config_, server);
 
     static constexpr auto IOQUEUE_MAX_HANDLES = min(PJ_IOQUEUE_MAX_HANDLES, 64);
     TRY( pj_timer_heap_create(pool_.get(), 100, &config_.stun_cfg.timer_heap) );
@@ -396,56 +440,48 @@ IceTransport::start(const Attribute& rem_attrs,
     return true;
 }
 
-std::string
-IceTransport::unpackLine(std::vector<uint8_t>::const_iterator& begin,
-                         std::vector<uint8_t>::const_iterator& end)
-{
-    if (std::distance(begin, end) <= 0)
-        return {};
-
-    // Search for EOL
-    std::vector<uint8_t>::const_iterator line_end(begin);
-    while (line_end != end && *line_end != NEW_LINE && *line_end)
-        ++line_end;
-
-    if (std::distance(begin, line_end) <= 0)
-        return {};
-
-    std::string str(begin, line_end);
-
-    // Consume the new line character
-    if (std::distance(line_end, end) > 0)
-        ++line_end;
-
-    begin = line_end;
-    return str;
-}
-
 bool
 IceTransport::start(const std::vector<uint8_t>& rem_data)
 {
-    auto begin = rem_data.cbegin();
-    auto end = rem_data.cend();
-    auto rem_ufrag = unpackLine(begin, end);
-    auto rem_pwd = unpackLine(begin, end);
-    if (rem_pwd.empty() or rem_pwd.empty()) {
-        RING_ERR("ICE remote attributes parsing error");
-        return false;
-    }
+    std::string rem_ufrag;
+    std::string rem_pwd;
     std::vector<IceCandidate> rem_candidates;
+
+    auto data = reinterpret_cast<const char*>(rem_data.data());
+    auto size = rem_data.size();
+
     try {
-        while (true) {
-            IceCandidate candidate;
-            const auto line = unpackLine(begin, end);
-            if (line.empty())
-                break;
-            if (getCandidateFromSDP(line, candidate))
-                rem_candidates.push_back(candidate);
+        std::size_t offset = 0;
+        auto result = msgpack::unpack(data, size, offset);
+        auto version = result.get().as<uint8_t>();
+        RING_DBG("[ice:%p] rx msg v%u", this, version);
+        if (version == 1) {
+            result = msgpack::unpack(data, size, offset);
+            std::tie(rem_ufrag, rem_pwd) = result.get().as<std::pair<std::string, std::string>>();
+            result = msgpack::unpack(data, size, offset);
+            auto comp_cnt = result.get().as<uint8_t>();
+            while (comp_cnt-- > 0) {
+                result = msgpack::unpack(data, size, offset);
+                IceCandidate cand;
+                for (const auto& line : result.get().as<std::vector<std::string>>()) {
+                    if (getCandidateFromSDP(line, cand))
+                        rem_candidates.emplace_back(std::move(cand));
+                }
+            }
+        } else {
+            RING_ERR("[ice:%p] invalid msg version", this);
+            return false;
         }
-    } catch (std::exception& e) {
-        RING_ERR("ICE remote candidates parsing error");
+    } catch (const msgpack::unpack_error& e) {
+        RING_ERR("[ice:%p] remote msg unpack error: %s", this, e.what());
         return false;
     }
+
+    if (rem_pwd.empty() or rem_pwd.empty() or rem_candidates.empty()) {
+        RING_ERR("[ice:%p] invalid remote attributes", this);
+        return false;
+    }
+
     return start({rem_ufrag, rem_pwd}, rem_candidates);
 }
 
@@ -592,10 +628,6 @@ IceTransport::registerPublicIP(unsigned compId, const IpAddr& publicIP)
     if (not isInitialized())
         return false;
 
-    // Register only if no NAT traversal methods exists
-    if (upnp_ or config_.stun.port > 0 or config_.turn.port > 0)
-        return false;
-
     // Find the local candidate corresponding to local host,
     // then register a rflx candidate using given public address
     // and this local address as base. It's port is used for both address
@@ -619,13 +651,54 @@ IceTransport::registerPublicIP(unsigned compId, const IpAddr& publicIP)
 void
 IceTransport::addReflectiveCandidate(int comp_id, const IpAddr& base, const IpAddr& addr)
 {
+    // HACK-HACK-HACK-HACK-HACK-HACK-HACK-HACK-HACK-HACK-HACK-HACK-HACK-HACK-HACK-HACK-HACK-HACK
+    // WARNING: following implementation is a HACK of PJNATH !!
+    // ice_strans doesn't have any API that permit to inject ICE any kind of candidates.
+    // So, the hack consists in accessing hidden ICE session using a patched PJPNATH
+    // library with a new API exposing this session (pj_ice_strans_get_ice_sess).
+    // Then call pj_ice_sess_add_cand() with a carfully forged candidate:
+    // the transport_id field uses an index in ICE transport STUN servers array
+    // corresponding to a STUN server with the same address familly.
+    // This implies we hope they'll not be modification of transport_id meaning in future
+    // and no conflics with the borrowed STUN config.
+    // HACK-HACK-HACK-HACK-HACK-HACK-HACK-HACK-HACK-HACK-HACK-HACK-HACK-HACK-HACK-HACK-HACK-HACK
+
+    // borrowed from pjproject/pjnath/ice_strans.c, modified to be C++11'ized.
+    static auto CREATE_TP_ID = [](pj_uint8_t type, pj_uint8_t idx) {
+        return (pj_uint8_t)((type << 6) | idx);
+    };
+    static constexpr int SRFLX_PREF = 65535;
+    static constexpr int TP_STUN = 1;
+
+    // find a compatible STUN host with same address familly, normally all system enabled
+    // host addresses are represented, so we expect to always found this host
+    int idx = -1;
+    auto af = addr.getFamily();
+    if (af == AF_UNSPEC) {
+        RING_ERR("[ice] Unable to add reflective IP %s: unknown addess familly",
+                 addr.toString().c_str());
+        return;
+    }
+
+    for (unsigned i=0; i < config_.stun_tp_cnt; ++i) {
+        if (config_.stun_tp[i].af == af) {
+            idx = i;
+            break;
+        }
+    }
+    if (idx < 0) {
+        RING_ERR("[ice] Unable to add reflective IP %s: no suitable local STUN host found",
+                 addr.toString().c_str());
+        return;
+    }
+
     pj_ice_sess_cand cand;
 
     cand.type = PJ_ICE_CAND_TYPE_SRFLX;
-    cand.status = PJ_EPENDING; /* not used */
+    cand.status = PJ_EPENDING; // not used
     cand.comp_id = comp_id;
-    cand.transport_id = 1; /* 1 = STUN */
-    cand.local_pref = 65535; /* host */
+    cand.transport_id = CREATE_TP_ID(TP_STUN, idx); // HACK!!
+    cand.local_pref = SRFLX_PREF; // reflective
     /* cand.foundation = ? */
     /* cand.prio = calculated by ice session */
     /* make base and addr the same since we're not going through a server */
@@ -694,19 +767,20 @@ IceTransport::selectUPnPIceCandidates()
 }
 
 std::vector<uint8_t>
-IceTransport::getLocalAttributesAndCandidates() const
+IceTransport::packIceMsg() const
 {
+    static constexpr uint8_t ICE_MSG_VERSION = 1;
+
     if (not isInitialized())
         return {};
 
     std::stringstream ss;
-    ss << local_ufrag_ << NEW_LINE;
-    ss << local_pwd_ << NEW_LINE;
-    for (unsigned i=0; i<component_count_; i++) {
-        const auto& candidates = getLocalCandidates(i);
-        for (const auto& c : candidates)
-            ss << c << NEW_LINE;
-    }
+    msgpack::pack(ss, ICE_MSG_VERSION);
+    msgpack::pack(ss, std::make_pair(local_ufrag_, local_pwd_));
+    msgpack::pack(ss, static_cast<uint8_t>(component_count_));
+    for (unsigned i=0; i<component_count_; i++)
+        msgpack::pack(ss, getLocalCandidates(i));
+
     auto str(ss.str());
     return std::vector<uint8_t>(str.begin(), str.end());
 }
@@ -892,6 +966,8 @@ IceTransport::waitForData(int comp_id, unsigned int timeout)
     return io.queue.front().datalen;
 }
 
+//##################################################################################################
+
 IceTransportFactory::IceTransportFactory()
     : cp_()
     , pool_(nullptr, pj_pool_release)
@@ -909,14 +985,13 @@ IceTransportFactory::IceTransportFactory()
     // v2.4.5 of PJNATH has a default of 100ms but RFC 5389 since version 14 requires
     // a minimum of 500ms on fixed-line links. Our usual case is wireless links.
     // This solves too long ICE exchange by DHT.
-    // Using 500ms with default PJ_STUN_MAX_TRANSMIT_COUNT (7) gives around 31s before timeout.
+    // Using 500ms with default PJ_STUN_MAX_TRANSMIT_COUNT (7) gives around 33s before timeout.
     ice_cfg_.stun_cfg.rto_msec = 500;
 
-    ice_cfg_.af = pj_AF_INET();
+    // Add local hosts (IPv4, IPv6) as stun candidates
+    add_stun_server(ice_cfg_, pj_AF_INET6());
+    add_stun_server(ice_cfg_, pj_AF_INET());
 
-    ice_cfg_.stun.cfg.max_pkt_size = 8192;
-    ice_cfg_.turn.cfg.max_pkt_size = 8192;
-    //ice_cfg_.stun.max_host_cands = icedemo.opt.max_host;
     ice_cfg_.opt.aggressive = PJ_FALSE;
 }
 
