@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2004-2016 Savoir-faire Linux Inc.
+ *  Copyright (C) 2004-2017 Savoir-faire Linux Inc.
  *
  *  Author: Alexandre Bourget <alexandre.bourget@savoirfairelinux.com>
  *  Author: Yan Morin <yan.morin@savoirfairelinux.com>
@@ -34,6 +34,7 @@
 #include "manager.h"
 #include "account_schema.h"
 #include "plugin_manager.h"
+#include "thread_pool.h"
 
 #include "fileutils.h"
 #include "map_utils.h"
@@ -133,6 +134,7 @@ static constexpr const char* DHTLOGLEVEL = "DHTLOGLEVEL";
 static void
 setDhtLogLevel()
 {
+#ifndef RING_UWP
     char* envvar = getenv(DHTLOGLEVEL);
     int level = 0;
 
@@ -145,6 +147,9 @@ setDhtLogLevel()
         RING_DBG("DHTLOGLEVEL=%u", level);
     }
     Manager::instance().dhtLogLevel = level;
+#else
+    Manager::instance().dhtLogLevel = 0;
+#endif
 }
 
 /**
@@ -159,7 +164,9 @@ static constexpr const char* SIPLOGLEVEL = "SIPLOGLEVEL";
 static void
 setSipLogLevel()
 {
+#ifndef RING_UWP
     char* envvar = getenv(SIPLOGLEVEL);
+
     int level = 0;
 
     if (envvar != nullptr) {
@@ -171,6 +178,9 @@ setSipLogLevel()
     }
 
     pj_log_set_level(level);
+#else
+    pj_log_set_level(0);
+#endif
 }
 
 /**
@@ -190,6 +200,7 @@ tls_print_logs(int level, const char* msg)
 static void
 setGnuTlsLogLevel()
 {
+#ifndef RING_UWP
     char* envvar = getenv("RING_TLS_LOGLEVEL");
     int level = RING_TLS_LOGLEVEL;
 
@@ -203,6 +214,9 @@ setGnuTlsLogLevel()
     }
 
     gnutls_global_set_log_level(level);
+#else
+    gnutls_global_set_log_level(RING_TLS_LOGLEVEL);
+#endif
     gnutls_global_set_log_function(tls_print_logs);
 }
 
@@ -394,6 +408,7 @@ Manager::finish() noexcept
 
         ice_tf_.reset();
         pj_shutdown();
+        ThreadPool::instance().join();
     } catch (const VoipLinkException &err) {
         RING_ERR("%s", err.what());
     }
@@ -438,7 +453,7 @@ Manager::switchCall(const std::string& id)
 }
 
 void
-Manager::switchCall(std::shared_ptr<Call> call)
+Manager::switchCall(const std::shared_ptr<Call>& call)
 {
     switchCall(call->getCallId());
 }
@@ -1430,7 +1445,7 @@ Manager::scheduleTask(const std::function<void()>&& task, std::chrono::steady_cl
 
 
 void
-Manager::scheduleTask(std::shared_ptr<Runnable> task, std::chrono::steady_clock::time_point when)
+Manager::scheduleTask(const std::shared_ptr<Runnable>& task, std::chrono::steady_clock::time_point when)
 {
     std::lock_guard<std::mutex> lock(scheduledTasksMutex_);
     scheduledTasks_.emplace(when, task);
@@ -2383,6 +2398,8 @@ Manager::setAccountsOrder(const std::string& order)
     preferences.setAccountOrder(order);
 
     saveConfig();
+
+    emitSignal<DRing::ConfigurationSignal::AccountsChanged>();
 }
 
 std::vector<std::string>
