@@ -1,7 +1,8 @@
 /*
- *  Copyright (C) 2014-2016 Savoir-faire Linux Inc.
+ *  Copyright (C) 2014-2017 Savoir-faire Linux Inc.
  *
  *  Author: Adrien Béraud <adrien.beraud@savoirfairelinux.com>
+ *  Author: Simon Désaulniers <simon.desaulniers@gmail.com>
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -83,6 +84,7 @@ constexpr const char* const RING_ACCOUNT_KEY = "ringAccountKey";
 constexpr const char* const RING_ACCOUNT_CERT = "ringAccountCert";
 constexpr const char* const RING_ACCOUNT_RECEIPT = "ringAccountReceipt";
 constexpr const char* const RING_ACCOUNT_RECEIPT_SIG = "ringAccountReceiptSignature";
+constexpr const char* const RING_ACCOUNT_CRL = "ringAccountCRL";
 }
 
 class IceTransport;
@@ -100,6 +102,13 @@ class RingAccount : public SIPAccountBase {
             return ACCOUNT_TYPE;
         }
 
+        std::shared_ptr<RingAccount> shared() {
+            return std::static_pointer_cast<RingAccount>(shared_from_this());
+        }
+        std::shared_ptr<RingAccount const> shared() const {
+            return std::static_pointer_cast<RingAccount const>(shared_from_this());
+        }
+
         /**
          * Constructor
          * @param accountID The account identifier
@@ -138,6 +147,22 @@ class RingAccount : public SIPAccountBase {
          */
         void loadConfig() override {}
 
+        /**
+         * Adds an account id to the list of accounts to track on the DHT for
+         * buddy presence.
+         *
+         * @param buddy_id  The buddy id.
+         */
+        void trackBuddyPresence(const std::string& buddy_id);
+
+        /**
+         * Tells for each tracked account id if it has been seen online so far
+         * in the last DeviceAnnouncement::TYPE.expiration minutes.
+         *
+         * @return map of buddy_uri to bool (online or not)
+         */
+        std::map<std::string, bool> getTrackedBuddyPresence();
+
         /**
          * Connect to the DHT.
          */
@@ -203,18 +228,6 @@ class RingAccount : public SIPAccountBase {
         /* Returns true if the username and/or hostname match this account */
         MatchRank matches(const std::string &username, const std::string &hostname) const override;
 
-        /**
-         * Activate the module.
-         * @param function Publish or subscribe to enable
-         * @param enable Flag
-         */
-        void enablePresence(const bool& enable);
-        /**
-         * Activate the publish/subscribe.
-         * @param enable Flag
-         */
-        void supportPresence(int function, bool enable);
-
         /**
          * Implementation of Account::newOutgoingCall()
          * Note: keep declaration before newOutgoingCall template.
@@ -228,9 +241,15 @@ class RingAccount : public SIPAccountBase {
          *      The type of this instance is given in template argument.
          *      This type can be any base class of SIPCall class (included).
          */
+#ifndef RING_UWP
         template <class T=SIPCall>
         std::shared_ptr<enable_if_base_of<T, SIPCall> >
         newOutgoingCall(const std::string& toUrl);
+#else
+        template <class T>
+        std::shared_ptr<T>
+        newOutgoingCall(const std::string& toUrl);
+#endif
 
         /**
          * Create incoming SIPCall.
@@ -264,7 +283,7 @@ class RingAccount : public SIPAccountBase {
         std::vector<std::string> getCertificatesByStatus(tls::TrustStore::PermissionStatus status);
 
         bool findCertificate(const std::string& id);
-        bool findCertificate(const dht::InfoHash& h, std::function<void(const std::shared_ptr<dht::crypto::Certificate>)> cb = {});
+        bool findCertificate(const dht::InfoHash& h, std::function<void(const std::shared_ptr<dht::crypto::Certificate>&)>&& cb = {});
 
         /* contact requests */
         std::map<std::string, std::string> getTrustRequests() const;
@@ -276,15 +295,9 @@ class RingAccount : public SIPAccountBase {
 
         void addDevice(const std::string& password);
 
-        std::map<std::string, std::string> getKnownDevices() const {
-            std::map<std::string, std::string> ids;
-            for (auto& d : knownDevices_) {
-                auto id = d.first.toString();
-                auto label = id.substr(0, 8);
-                ids.emplace(std::move(id), std::move(label));
-            }
-            return ids;
-        }
+        bool revokeDevice(const std::string& password, const std::string& device);
+
+        std::map<std::string, std::string> getKnownDevices() const;
 
         void connectivityChanged() override;
 
@@ -300,56 +313,21 @@ class RingAccount : public SIPAccountBase {
     private:
         NON_COPYABLE(RingAccount);
 
-        struct PendingCall {
-            std::chrono::steady_clock::time_point start;
-            std::shared_ptr<IceTransport> ice_sp;
-            std::weak_ptr<SIPCall> call;
-            std::future<size_t> listen_key;
-            dht::InfoHash call_key;
-            dht::InfoHash from;
-            std::shared_ptr<dht::crypto::Certificate> from_cert;
-        };
-
-        struct PendingMessage {
-            dht::InfoHash to;
-            std::chrono::steady_clock::time_point received;
-        };
-
-        struct TrustRequest {
-            dht::InfoHash from;
-            std::chrono::system_clock::time_point received;
-            std::vector<uint8_t> payload;
-        };
-
         /**
-         * Crypto material contained in the archive,
-         * not persisted in the account configuration
+         * Private structures
          */
-        struct ArchiveContent {
-            /** Account main private key and certificate chain */
-            dht::crypto::Identity id;
+        struct PendingCall;
+        struct PendingMessage;
+        struct SavedTrustRequest;
+        struct TrustRequest;
+        struct KnownDevice;
+        struct ArchiveContent;
+        struct DeviceAnnouncement;
+        struct DeviceSync;
+        struct BuddyInfo;
 
-            /** Generated CA key (for self-signed certificates) */
-            dht::crypto::PrivateKey ca_key;
-
-            /** Ethereum private key */
-            std::vector<uint8_t> eth_key;
-
-            /** Account configuration */
-            std::map<std::string, std::string> config;
-        };
-
-        /**
-         * Device announcement stored on DHT.
-         */
-        struct DeviceAnnouncement : public dht::SignedValue<DeviceAnnouncement> {
-        private:
-            using BaseClass = dht::SignedValue<DeviceAnnouncement>;
-        public:
-            static const constexpr dht::ValueType& TYPE = dht::ValueType::USER_DATA;
-            dht::InfoHash dev;
-            MSGPACK_DEFINE_MAP(dev);
-        };
+        void syncDevices();
+        void onReceiveDeviceSync(DeviceSync&& sync);
 
 #if HAVE_RINGNS
         std::reference_wrapper<NameDirectory> nameDir_;
@@ -362,13 +340,30 @@ class RingAccount : public SIPAccountBase {
          */
         static std::pair<std::vector<uint8_t>, dht::InfoHash> computeKeys(const std::string& password, const std::string& pin, bool previous=false);
 
+        /**
+         * Update tracking info when buddy appears offline.
+         *
+         * @param buddy_info_it  An iterator over the map trackedBuddies_
+         */
+        void onTrackedBuddyOffline(std::map<dht::InfoHash, BuddyInfo>::iterator& buddy_info_it);
+
+        /**
+         * Update tracking info when buddy appears offline.
+         *
+         * @param buddy_info_it  An iterator over the map trackedBuddies_
+         * @param device_id       The device id
+         */
+        void onTrackedBuddyOnline(std::map<dht::InfoHash, BuddyInfo>::iterator& buddy_info_it, const dht::InfoHash& device_id);
+
         void doRegister_();
-        void incomingCall(dht::IceCandidates&& msg, std::shared_ptr<dht::crypto::Certificate> from);
+        void incomingCall(dht::IceCandidates&& msg, const std::shared_ptr<dht::crypto::Certificate>& from);
 
         const dht::ValueType USER_PROFILE_TYPE = {9, "User profile", std::chrono::hours(24 * 7)};
 
         void handleEvents();
 
+        void forEachDevice(const dht::InfoHash& to, std::function<void(const std::shared_ptr<RingAccount>&, const dht::InfoHash&)> op, std::function<void(bool)> end = {});
+
         void createOutgoingCall(const std::shared_ptr<SIPCall>& call, const std::string& to_id, IpAddr target);
 
         /**
@@ -377,7 +372,7 @@ class RingAccount : public SIPAccountBase {
          */
         virtual void setAccountDetails(const std::map<std::string, std::string> &details) override;
 
-        void startOutgoingCall(std::shared_ptr<SIPCall>& call, const std::string toUri);
+        void startOutgoingCall(const std::shared_ptr<SIPCall>& call, const std::string toUri);
 
         /**
          * Start a SIP Call
@@ -386,7 +381,26 @@ class RingAccount : public SIPAccountBase {
          */
         bool SIPStartCall(const std::shared_ptr<SIPCall>& call, IpAddr target);
 
-        bool foundKnownDevice(const std::shared_ptr<dht::crypto::Certificate>& crt);
+        /**
+         * Inform that a potential account device have been found.
+         * Returns true if the device have been validated to be part of this account
+         */
+        bool foundAccountDevice(const std::shared_ptr<dht::crypto::Certificate>& crt, const std::string& name = {});
+
+        /**
+         * Inform that a potential peer device have been found.
+         * Returns true only if the device certificate is a valid Ring device certificate.
+         * In that case (true is returned) the account_id parameter is set to the peer account ID.
+         */
+        bool foundPeerDevice(const std::shared_ptr<dht::crypto::Certificate>& crt, dht::InfoHash& account_id);
+
+        /**
+         * Check that a peer is authorised to talk to us.
+         * If everything is in order, calls the callback with the
+         * peer certificate chain (down to the peer device certificate),
+         * and the peer account id.
+         */
+        void onPeerMessage(const dht::InfoHash& peer_device, std::function<void(const std::shared_ptr<dht::crypto::Certificate>& crt, const dht::InfoHash& account_id)>);
 
         /**
          * Maps require port via UPnP
@@ -406,20 +420,21 @@ class RingAccount : public SIPAccountBase {
         /**
          * DHT calls waiting for ICE negotiation
          */
-        std::list<PendingCall> pendingCalls_ {};
+        std::list<PendingCall> pendingCalls_;
 
         /**
          * Incoming DHT calls that are not yet actual SIP calls.
          */
-        std::list<PendingCall> pendingSipCalls_ {};
+        std::list<PendingCall> pendingSipCalls_;
         std::set<dht::Value::Id> treatedCalls_ {};
         mutable std::mutex callsMutex_ {};
 
-        std::map<dht::Value::Id, PendingMessage> sentMessages_ {};
+        std::map<dht::Value::Id, PendingMessage> sentMessages_;
         std::set<dht::Value::Id> treatedMessages_ {};
 
         std::string ringAccountId_ {};
         std::string ringDeviceId_ {};
+        std::string ringDeviceName_ {};
         std::string idPath_ {};
         std::string cachePath_ {};
         std::string dataPath_ {};
@@ -432,12 +447,20 @@ class RingAccount : public SIPAccountBase {
         std::vector<uint8_t> receiptSignature_ {};
         dht::Value announceVal_;
 
-        std::vector<TrustRequest> trustRequests_;
+        std::map<dht::InfoHash, TrustRequest> trustRequests_;
+        void loadTrustRequests();
+        void saveTrustRequests();
 
         tls::TrustStore trust_;
 
         std::shared_ptr<dht::Value> announce_;
-        std::map<dht::InfoHash, std::shared_ptr<dht::crypto::Certificate>> knownDevices_;
+
+        /* this ring account associated devices */
+        std::map<dht::InfoHash, KnownDevice> knownDevices_;
+
+        /* tracked buddies presence */
+        std::recursive_mutex buddyInfoMtx;
+        std::map<dht::InfoHash, BuddyInfo> trackedBuddies_;
 
         void loadAccount(const std::string& archive_password = {}, const std::string& archive_pin = {});
         void loadAccountFromDHT(const std::string& archive_password, const std::string& archive_pin);
@@ -445,10 +468,13 @@ class RingAccount : public SIPAccountBase {
         bool hasCertificate() const;
         bool hasPrivateKey() const;
         bool hasSignedReceipt();
+        bool needsMigration() const;
 
         std::string makeReceipt(const dht::crypto::Identity& id);
         void createRingDevice(const dht::crypto::Identity& id);
         void initRingDevice(const ArchiveContent& a);
+        bool migrateAccount(const std::string& pwd);
+        static bool updateCertificates(ArchiveContent& archive, dht::crypto::Identity& device);
 
         void createAccount(const std::string& archive_password);
         std::vector<uint8_t> makeArchive(const ArchiveContent& content) const;
@@ -457,7 +483,7 @@ class RingAccount : public SIPAccountBase {
         static ArchiveContent loadArchive(const std::vector<uint8_t>& data);
         std::vector<std::pair<sockaddr_storage, socklen_t>> loadBootstrap() const;
 
-        void saveIdentity(const dht::crypto::Identity id, const std::string& path) const;
+        std::pair<std::string, std::string> saveIdentity(const dht::crypto::Identity id, const std::string& path) const;
         void saveNodes(const std::vector<dht::NodeExport>&) const;
         void saveValues(const std::vector<dht::ValuesExport>&) const;
 
@@ -467,13 +493,14 @@ class RingAccount : public SIPAccountBase {
         void loadTreatedMessages();
         void saveTreatedMessages() const;
 
+        void loadKnownDevicesOld();
         void loadKnownDevices();
         void saveKnownDevices() const;
 
-        void replyToIncomingIceMsg(std::shared_ptr<SIPCall>,
-                                   std::shared_ptr<IceTransport>,
+        void replyToIncomingIceMsg(const std::shared_ptr<SIPCall>&,
+                                   const std::shared_ptr<IceTransport>&,
                                    const dht::IceCandidates&,
-                                   std::shared_ptr<dht::crypto::Certificate>);
+                                   const std::shared_ptr<dht::crypto::Certificate>&);
 
         static tls::DhParams loadDhParams(const std::string path);
 
@@ -527,7 +554,7 @@ class RingAccount : public SIPAccountBase {
         /**
          * Optional: via_addr construct from received parameters
          */
-        pjsip_host_port via_addr_;
+        pjsip_host_port via_addr_ {};
 
         char contactBuffer_[PJSIP_MAX_URL_SIZE] {};
         pj_str_t contact_ {contactBuffer_, 0};
